from .. import position_paradigm
from .. import hyperoctahedral

def test_hyperoctahedral_group_irreps():
    for n in (2,3,4,5):
        framework = position_paradigm.GenomeFramework(n)
        for irrep in framework.irreps():
            r1 = framework.random_instance()
            r2 = framework.random_instance()
            assert (irrep(r1) * irrep(r2)) == irrep(r2*r1)

def test_hyperoctahedral_group_map():
    for n in (3, 4, 5, 10, 20, 30):
        framework = position_paradigm.GenomeFramework(n)
        H = hyperoctahedral.HyperoctahedralGroup(n)
        for _ in range(10):
            r1 = framework.random_instance()
            r2 = framework.random_instance()
            assert H.Phi(r1) * H.Phi(r2) == H.Phi(r1*r2)




def newick_to_json(
    newick_string,
    namestring="id",
    lengthstring="branch_length",
    childrenstring="children",
    generate_names=False,
):
    """Parses a newick string (example "((A,B),C)") into JSON format accepted by NetworkX

    Arguments:
            newick_string: The newick string to convert to JSON. Names, branch lengths and named
                    internal nodes accepted. If no names are given, an ID is generated for the name if generate_names=True.
            namestring: The label for node name to use in the JSON (default "id").
            lengthstring: The label for branch length to use in the JSON (default "branch_length").
            childrenstring: The label for children arrays to use in the JSON (default "children").
            generate_names: Whether or not to generate names if none are given
    """
    newick_string = newick_string.strip(";")
    node = dict()
    info = newick_string[newick_string.rfind(")") + 1:]
    info = info.split(":")
    name = info[0]
    node[namestring] = name
    length = ""
    if len(info) > 1:
        length = info[1]
        node[lengthstring] = float(length)
    children_string = newick_string[0 : newick_string.rfind(")") + 1]
    if children_string != "":
        if children_string[-1] == ")":
            children_string = children_string[1:-1]
        children_string = __split_into_children(children_string)
        child_nodes_JSON = []
        generated_name_parts = []
        for child_string in children_string:
            child_JSON = newick_to_json(
                child_string, namestring, lengthstring, childrenstring, generate_names
            )
            if (not name) and generate_names:
                generated_name_parts.append(child_JSON[namestring])
            child_nodes_JSON.append(child_JSON)
        if generate_names:
            node[namestring] = "|".join(generated_name_parts)
        node[childrenstring] = child_nodes_JSON
    return node


def json_to_newick(json_dict):
    """Converts a JSON tree to a newick string

    Arguments:
            json_dict: The JSON tree to convert to newick.
    """
    return __json_to_newick(json_dict) + ";"


def __json_to_newick(
    json_dict, namestring="id", lengthstring="branch_length", childrenstring="children"
):
    if not isinstance(json_dict, list):
        string = ""
        if childrenstring in json_dict:
            string += (
                "("
                + __json_to_newick(json_dict[childrenstring])
                + ")"
                + json_dict[namestring]
            )
        else:
            string += json_dict[namestring]
        if lengthstring in json_dict:
            string += ":" + str(json_dict[lengthstring])
        return string
    else:
        string = ""
        for dict in json_dict:
            string += __json_to_newick(dict) + ","
        return string[0:-1]


def __split_into_children(children_string):
    """Helper function for splitting a newick string into siblings"""
    string = children_string
    opened = 0
    closed = 0
    pieces = []
    last_stop = 0
    for i, s in enumerate(string):
        if s == "(":
            opened += 1
        if s == ")":
            closed += 1
        if s == "," and opened == closed:
            pieces.append(string[last_stop:i])
            last_stop = i + 1
    pieces.append(string[last_stop:])
    return pieces



"""
"""

import os
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from .position_paradigm import *
from .enums import *


def draw_genome_instance(framework, instance, show=False):
    """Produce a drawing of a circular genome with n regions"""
    if not framework.oriented:
        raise NotImplementedError("not implemented for linear genomes.")
    try:
        instance = framework.cycles(instance)
    except:
        TypeError(f"{instance} does not belong in {str(framework)}")

    matplotlib.rcParams.update(
        {
            "text.usetex": True,
            "font.family": "serif",
            "pgf.texsystem": "pdflatex",
            "pgf.rcfonts": False,
            "font.size": 18,
        }
    )
    plt.style.use("default")

    n = framework.n
    region_labels = [None for _ in range(1, n + 1)]
    orientation_list = [False for _ in range(1, n + 1)]

    for r, p in list(enumerate(framework.one_row(instance, as_list=True))):
        region_labels[abs(p) - 1] = r + 1
        orientation_list[abs(p) - 1] = p < 0

    region_labels = list(reversed(region_labels))
    orientation_list = list(reversed(orientation_list))

    segment_sizes = [100 / n for _ in range(n)]
    segment_colors = ["white" for _ in range(n)]

    fig1, ax1 = plt.subplots(figsize=(2, 2))
    wedges, texts = ax1.pie(
        segment_sizes,
        labeldistance=1.2,
        colors=segment_colors,
        labels=region_labels,
        radius=1,
        startangle=90,
        wedgeprops={
            "edgecolor": "0",
            "linewidth": 1,
            "linestyle": "solid",
            "antialiased": True,
        },
    )

    if framework.oriented and orientation_list:
        for i, p in enumerate(wedges):
            ang = (p.theta2 - p.theta1) / 2.0 + p.theta1
            y = np.sin(np.deg2rad(ang))
            x = np.cos(np.deg2rad(ang))
            arrow_length = 0.00001  # We only want the arrow head
            angle = np.deg2rad(ang) + (
                np.pi / 2 if orientation_list[i] else 3 * np.pi / 2
            )
            if orientation_list:
                ax1.arrow(
                    x,
                    y,
                    arrow_length * np.cos(angle),
                    arrow_length * np.sin(angle),
                    overhang=0.3,
                    head_width=0.12,
                    head_length=0.12,
                    head_starts_at_zero=True,
                    length_includes_head=True,
                    linewidth=1,
                    facecolor="black",
                    fill=True,
                )

    # Draw inner circle to hide most of the segment edges
    centre_circle = plt.Circle((0, 0), 0.85, fc="white")
    fig = plt.gcf()
    fig.gca().add_artist(centre_circle)
    ax1.axis("equal")  # makes sure it's a circle
    plt.tight_layout()

    if not os.path.exists("_output"):
        os.makedirs("_output")
    plt.savefig(f"_output/{str(instance)}.png", transparent=True)
    plt.savefig(f"_output/{str(instance)}.pdf", transparent=True)
    if show:
        plt.show()



import pickle
import os

def retrieve_irrep_of_z(n=None):
    """
    Retrieve the irreps of z
    """
    dir_path = os.path.dirname(os.path.realpath(__file__)) + "/pickles/"
    full_path = dir_path + f"irreps_of_z_n{n}.pickle"
    try:
        with open(full_path, 'rb') as handle:
            irreps_of_z = pickle.load(handle)
    except FileNotFoundError:
        print(f"File not found: irreps_of_z_n{n}.pickle")
        return None
    
    return irreps_of_z


"""
This module contains functions for producing a selection of pre-styled matplotlib plots, with settings specific to the context of circular genomes with minimal other options available. For fully custom plots, use the matplotlib library directly along with the cgt.distances module.
"""

import cgt
import matplotlib as mpl
import matplotlib.pyplot as plt
from cycler import cycler
import numpy as np
from cgt.enums import DATA


def latex_figure_setup():
    """
    Sets up the matplotlib environment to use LaTeX for all text.
    """
    nice_fonts = {
        # Use LaTeX to write all text
        "text.usetex": True,
        "font.family": "serif",
        "axes.labelsize": 11,
        "font.size": 11,
        "legend.fontsize": 11,
        "xtick.labelsize": 11,
        "ytick.labelsize": 11,
    }
    plt.style.use("default")
    mpl.rcParams.update(nice_fonts)
    monochrome = (
        cycler("color", ["k"])
        * cycler("marker", ["", "."])
        * cycler("linestyle", ["-", "--", ":"])
    )
    plt.rc("axes", prop_cycle=monochrome)


def x_axis_limit_for_framework(framework):
    n = framework.n
    from_testing = { 1: 10, 2: 10, 3: 10, 4: 14, 5: 22, 6: 26, 7: 34, 8: 40, 9: 98 }
    if n in from_testing:
        return from_testing[n]
    else:
        return from_testing[2 * max(from_testing.keys())]


def plot_likelihood_and_probability(framework, model, instance, width=6, height=4, tmax = None, dpi=300, use_eigenvectors=True):
    canonical_instance = framework.canonical_instance(instance)
    L = cgt.distances.likelihood_function(
        framework, model, instance, use_eigenvectors=use_eigenvectors
    )
    if tmax is None:
        max_x = x_axis_limit_for_framework(framework)
    else:
        max_x = tmax
    if DATA.reg_rep_of_zs in model.data_bundle:
        stepwise_dists = cgt.distances.fast_step_probabilities(framework, model, limit = max_x + 1)[canonical_instance]
        def a(k):
            return stepwise_dists[k]
    else:
        a = cgt.distances.prob_to_reach_in_steps_func(framework, model, instance, use_eigenvectors=use_eigenvectors)
    tvec_L = list(np.arange(0, max_x + 1, 0.2))
    tvec = list(np.arange(0, max_x + 1, 1))
    latex_figure_setup()
    plt.figure(figsize=(width, height), dpi=dpi)
    plt.scatter(tvec, [a(t) for t in tvec], color="black", s=3)
    plt.plot(
        tvec,
        [a(t) for t in tvec],
        color="black",
        linestyle="dotted",
        label="Probability",
        linewidth=0.9,
    )
    plt.plot(
        tvec_L,
        [L(t) for t in tvec_L],
        color="black",
        linestyle="solid",
        label="Likelihood",
    )
    plt.title(f"Genome: z{str(canonical_instance)}")
    n = framework.n
    if n < 9:
        x_tick_list = list(range(0, 9, 1)) + list(range(10, max_x + 1, 2))
    else: 
        x_tick_list = list(range(0, max_x + 1, 5))
    plt.xticks(x_tick_list)
    # Legend without border
    # bbox is (x, y, width, height)
    plt.legend(loc="best", frameon=False, bbox_to_anchor=(0.7, 0.0, 0.3, 1.0))
    # As the x axis label, show the MLE
    mle = cgt.distances.maximise(framework, L)
    plt.xlabel(
        f"MLE: {round(mle, 1) if not np.isnan(mle) else 'N/A'}, Min dist: {cgt.distances.first_nonzero_value(framework, a)}"
    )
    plt.show()
    return a, L



from .. import example
import pytest

@pytest.fixture
def example_framework():
    yield example(10)

def test_one_row_and_cycles_random(example_framework):
    fw, _ = example_framework
    instance = fw.random_instance()
    assert instance == fw.cycles(fw.one_row(instance))

def test_same_canonical_instance_random(example_framework):
    fw, _ = example_framework
    canonical_instance = fw.canonical_instance(fw.random_instance())
    genome = fw._genome_coset(fw.cycles(canonical_instance))
    for instance in genome:
        assert canonical_instance == fw.canonical_instance(instance)